SystemOrganization addCategory: #'PetitAnalyzer-Core'!
SystemOrganization addCategory: #'PetitAnalyzer-Tests'!

!PPStringParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:05'!
isTerminal
	^ true! !

!PPPluggableParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:05'!
isTerminal
	^ true! !

!PPEpsilonParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:11'!
isNullable
	^ true! !

!PPEpsilonParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:05'!
isTerminal
	^ true! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:59'!
firstSet
	"Answer the first-set of the receiver. Note, this implementation is inefficient when called on different receivers of the same grammar, instead use #firstSets to calculate the first-sets at once."
	
	^ self firstSets at: self! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:34'!
firstSets
	"Answer a dictionary with all the parsers reachable from the receiver as key and their first-set as value. The first-set of a parser is the list of terminal parsers that begin the parser derivable from that parser."
	
	| firstSets |
	firstSets := IdentityDictionary new.
	self allNodesDo: [ :each | 
		firstSets at: each put: (each isTerminal
			ifTrue: [ IdentitySet with: each ]
			ifFalse: [ IdentitySet new ]) ].
	[	| changed tally |
		changed := false.
		firstSets keysAndValuesDo: [ :parser :first |
			tally := first size.
			parser firstSets: firstSets into: first.
			changed := changed or: [ tally ~= first size ] ].
		changed ] whileTrue.
	^ firstSets! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:58'!
firstSets: aFirstDictionary into: aSet
	"PROTECTED: Try to add additional elements to the first-set aSet of the receiver, use the incomplete aFirstDictionary."

	self children do: [ :parser | aSet addAll: (aFirstDictionary at: parser) ]! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:57'!
followSet
	"Answer the follow-set of the receiver starting at the receiver. Note, this implementation is inefficient when called on different receivers of the same grammar, instead use #followSets to calculate the follow-sets at once."
	
	^ self followSetIn: self! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:56'!
followSetIn: aParser
	"Answer the follow-set of the receiver starting at aParser. Note, this implementation is inefficient when called on different receivers of the same grammar, instead use #followSets to calculate the follow-sets at once."
	
	^ aParser followSets at: self! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:57'!
followSets
	"Answer a dictionary with all the parsers reachable from the receiver as key and their follow-set as value. The follow-set of a parser is the list of terminal parsers that can appear immediately to the right of that parser."
	
	| current previous continue firstSets followSets |
	current := previous := 0.
	firstSets := self firstSets.
	followSets := IdentityDictionary new.
	self allNodesDo: [ :each | followSets at: each put: IdentitySet new ].
	(followSets at: self) add: PPEpsilonParser new.
	[	followSets keysAndValuesDo: [ :parser :follow |
			parser 
				followSets: followSets
				firstSets: firstSets
				into: follow ].
		current := followSets
			inject: 0
			into: [ :result :each | result + each size ].
		continue := previous < current.
		previous := current.
		continue ] whileTrue.
	^ followSets! !

!PPParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:57'!
followSets: aFollowDictionary firstSets: aFirstDictionary into: aSet
	"PROTECTED: Try to add additional elements to the follow-set aSet of the receiver, use the incomplete aFollowDictionary and the complete aFirstDictionary."
	
	self children do: [ :parser | (aFollowDictionary at: parser) addAll: aSet ]! !

!PPParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:11'!
isNullable
	"Answer true if the receiver is a nullable parser, it can successfully parse nothing."
	
	^ false! !

!PPParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 18:19'!
isTerminal
	"Answer true if the receiver is a terminal parser, that means it does not delegate to any other parser."
	
	^ false! !

!PPSequenceParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 18:45'!
firstSets: aFirstDictionary into: aSet
	| nullable |
	parsers do: [ :parser |
		nullable := false.
		(aFirstDictionary at: parser) do: [ :each |
			each isNullable
				ifTrue: [ nullable := true ]
				ifFalse: [ aSet add: each ] ].
		nullable
			ifFalse: [ ^ self ] ].
	aSet add: PPEpsilonParser new! !

!PPSequenceParser methodsFor: '*petitanalyzer-querying' stamp: 'lr 10/22/2009 19:39'!
followSets: aFollowDictionary firstSets: aFirstDictionary into: aSet
	parsers withIndexDo: [ :parser :index |
		| followSet firstSet |
		followSet := aFollowDictionary at: parser.
		index = parsers size
			ifTrue: [ followSet addAll: aSet ]
			ifFalse: [
				(self class withAll: (parsers 
					copyFrom: index + 1 to: parsers size))
						firstSets: aFirstDictionary
						into: (firstSet := IdentitySet new).
				(firstSet anySatisfy: [ :each | each isNullable ])
					ifTrue: [ followSet addAll: aSet ].
				followSet addAll: (firstSet 
					reject: [ :each | each isNullable ]) ] ]! !

Object subclass: #PPValidator
	instanceVariableNames: 'parser productions firstSets followSets'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PetitAnalyzer-Core'!

!PPValidator class methodsFor: 'instance-creation' stamp: 'lr 10/23/2009 11:50'!
on: aParser
	^ self basicNew initializeOn: aParser! !

!PPValidator methodsFor: 'initialization' stamp: 'lr 10/23/2009 11:59'!
initializeOn: aParser
	parser := aParser.
	productions := aParser allNodes.
	firstSets := aParser firstSets.
	followSets := aParser followSets! !

!PPValidator methodsFor: 'testing-LL(1)' stamp: 'lr 10/23/2009 19:29'!
validateFirstFirstConflict
	"FIRST/FIRST conflict: The FIRST sets of two different non-terminals are overlapping."
	
	| reverseSets |
	reverseSets := IdentityDictionary new.
	firstSets keysAndValuesDo: [ :production :firstSet |
		production isTerminal ifFalse: [ 
			firstSet do: [ :other |
				(reverseSets at: other ifAbsentPut: [ IdentitySet new ])
					add: production ] ] ].
	reverseSets keysAndValuesDo: [ :production :reverseSet |
		reverseSet size > 1 
			ifTrue: [ Transcript show: 'LL(1): First-First Conflict: '; show: reverseSet; cr ] ]! !

!PPValidator methodsFor: 'testing-LL(1)' stamp: 'lr 10/23/2009 19:29'!
validateFirstFollowConflict
	"FIRST/FOLLOW conflict: The FIRST and FOLLOW set of a grammar rule overlap. With an epsilon in the FIRST set it is unknown which alternative to select."
	
	| firstSet followSet |
	productions do: [ :production |
		firstSet := firstSets at: production.
		(firstSet anySatisfy: [ :each | each isNullable ]) ifTrue: [
			followSet := followSets at: production.
			(firstSet intersection: followSet) isEmpty
				ifFalse: [ Transcript show: 'LL(1): First-Follow Conflict: '; show: production; cr ] ] ]! !

PPAbstractParseTest subclass: #PPFirstFollowTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'PetitAnalyzer-Tests'!

!PPFirstFollowTest methodsFor: 'utilities' stamp: 'lr 10/22/2009 18:07'!
assert: aCollection includes: aString epsilon: aBoolean
	| parsers checker stream |
	parsers := aCollection
		collect: [ :each | each end ].
	checker := [ :string |
		parsers anySatisfy: [ :parser |
			(parser parse: string asParserStream)
				isFailure not ] ].
	stream := WriteStream on: String new.
	32 to: 127 do: [ :index |
		(checker value: (String with: (Character value: index)))
			ifTrue: [ stream nextPut: (Character value: index) ] ].
	self
		assert: stream contents = aString
		description: 'Expected ' , aString printString , ', but got ' , stream contents printString.
	self
		assert: (checker value: '') = aBoolean
		description: 'Expected epsilon to ' , (aBoolean ifTrue: [ 'be' ] ifFalse: [ 'not be' ]) , '  included'! !

!PPFirstFollowTest methodsFor: 'accessing' stamp: 'lr 10/21/2009 13:25'!
grammarA
	"Gting, Erwig, bersetzerbau, Springer (p.63)"

	| grammar |
	grammar := Dictionary new.
	
	" terminals "
	grammar at: #a put: $a asParser.
	grammar at: #b put: $b asParser.
	grammar at: #c put: $c asParser.
	grammar at: #d put: $d asParser.
	grammar at: #e put: PPEpsilonParser new.
	
	" non terminals "
	grammar at: #B put: (grammar at: #b) / (grammar at: #e).
	grammar at: #A put: (grammar at: #a) / (grammar at: #B).
	grammar at: #S put: (grammar at: #A) , (grammar at: #B) , (grammar at: #c) , (grammar at: #d).
	
	^ grammar	! !

!PPFirstFollowTest methodsFor: 'accessing' stamp: 'lr 10/22/2009 18:32'!
grammarB
	"The canonical grammar to exercise first- and follow-set calculation, probably originally from the dragon-book."

	| grammar |
	grammar := Dictionary new.
	#(E Ep T Tp F) 		do: [ :each | grammar at: each put: PPUnresolvedParser new ].
		
	(grammar at: #E)		def: (grammar at: #T) , (grammar at: #Ep).
	(grammar at: #Ep)	def: ($+ asParser , (grammar at: #T) , (grammar at: #Ep)) optional.
	
	(grammar at: #T)		def: (grammar at: #F) , (grammar at: #Tp).
	(grammar at: #Tp)	def: ($* asParser , (grammar at: #F) , (grammar at: #Tp)) optional.
	
	(grammar at: #F)		def: ($( asParser , (grammar at: #E) , $) asParser) / $i asParser.
			
	#(E Ep T Tp F) 		do: [ :each | (grammar at: each) name: each ].
	
	^ grammar! !

!PPFirstFollowTest methodsFor: 'accessing' stamp: 'lr 10/22/2009 18:32'!
grammarC
	"A highly recrusive grammar."

	| grammar |
	grammar := PPUnresolvedParser new.
	grammar def: (grammar , $+ asParser , grammar) / $1 asParser.
	^ grammar! !

!PPFirstFollowTest methodsFor: 'testing-firstset' stamp: 'lr 10/22/2009 18:08'!
testFirstSetExampleA
	| grammar |
	grammar := self grammarA.
	self assert: (grammar at: #a) firstSet includes: 'a' epsilon: false.
	self assert: (grammar at: #b) firstSet includes: 'b' epsilon: false.
	self assert: (grammar at: #c) firstSet includes: 'c' epsilon: false.
	self assert: (grammar at: #d) firstSet includes: 'd' epsilon: false.
	self assert: (grammar at: #e) firstSet includes: '' epsilon: true.
	self assert: (grammar at: #S) firstSet includes: 'abc' epsilon: false.
	self assert: (grammar at: #A) firstSet includes: 'ab' epsilon: true.
	self assert: (grammar at: #B) firstSet includes: 'b' epsilon: true! !

!PPFirstFollowTest methodsFor: 'testing-firstset' stamp: 'lr 10/22/2009 18:09'!
testFirstSetExampleB
	| grammar |
	grammar := self grammarB.
	self assert: (grammar at: #E) firstSet includes: '(i' epsilon: false.
	self assert: (grammar at: #Ep) firstSet includes: '+' epsilon: true.
	self assert: (grammar at: #T) firstSet includes: '(i' epsilon: false.
	self assert: (grammar at: #Tp) firstSet includes: '*' epsilon: true.
	self assert: (grammar at: #F) firstSet includes: '(i' epsilon: false! !

!PPFirstFollowTest methodsFor: 'testing-firstset' stamp: 'lr 10/22/2009 18:30'!
testFirstSetExampleC
	| grammar |
	grammar := self grammarC.
	self assert: grammar firstSet includes: '1' epsilon: false! !

!PPFirstFollowTest methodsFor: 'testing-firstset' stamp: 'lr 10/22/2009 18:10'!
testFirstSetExpression
	| grammar |
	grammar := PPExpressionParser new.
	self assert: grammar start firstSet includes: '(-0123456789' epsilon: false.
	self assert: grammar addition firstSet includes: '(-0123456789' epsilon: false.
	self assert: grammar factors firstSet includes:  '(-0123456789' epsilon: false.
	self assert: grammar multiplication firstSet includes:  '(-0123456789' epsilon: false.
	self assert: grammar number firstSet includes: '-0123456789' epsilon: false.
	self assert: grammar parentheses firstSet includes: '(' epsilon: false.
	self assert: grammar power firstSet includes: '(-0123456789' epsilon: false.
	self assert: grammar primary firstSet includes: '(-0123456789' epsilon: false.
	self assert: grammar terms firstSet includes: '(-0123456789' epsilon: false! !

!PPFirstFollowTest methodsFor: 'testing-firstset' stamp: 'lr 10/22/2009 18:10'!
testFirstSetLambda
	| grammar |
	grammar := PPLambdaParser new.
	self assert: grammar start firstSet includes: '(ABCDEFGHIJKLMNOPQRSTUVWXYZ\abcdefghijklmnopqrstuvwxyz' epsilon: false.
	self assert: grammar abstraction firstSet includes: '\' epsilon: false.
	self assert: grammar application firstSet includes: '(' epsilon: false.
	self assert: grammar expression firstSet includes: '(ABCDEFGHIJKLMNOPQRSTUVWXYZ\abcdefghijklmnopqrstuvwxyz' epsilon: false.
	self assert: grammar variable firstSet includes: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz' epsilon: false! !

!PPFirstFollowTest methodsFor: 'testing-followset' stamp: 'lr 10/22/2009 19:53'!
testFollowSetExampleA
	| grammar followSets |
	grammar := self grammarA.
	followSets := (grammar at: #S) followSets.
	self assert: (followSets at: (grammar at: #a)) includes: 'bc' epsilon: false.
	self assert: (followSets at: (grammar at: #b)) includes: 'bc' epsilon: false.
	self assert: (followSets at: (grammar at: #c)) includes: 'd' epsilon: false.
	self assert: (followSets at: (grammar at: #d)) includes: '' epsilon: true.
	self assert: (followSets at: (grammar at: #e)) includes: 'bc' epsilon: false.
	self assert: (followSets at: (grammar at: #S)) includes: '' epsilon: true.
	self assert: (followSets at: (grammar at: #A)) includes: 'bc' epsilon: false.
	self assert: (followSets at: (grammar at: #B)) includes: 'bc' epsilon: false! !

!PPFirstFollowTest methodsFor: 'testing-followset' stamp: 'lr 10/22/2009 19:06'!
testFollowSetExampleB
	| grammar followSets |
	grammar := self grammarB.
	followSets := (grammar at: #E) followSets.
	self assert: (followSets at: (grammar at: #E)) includes: ')' epsilon: true.
	self assert: (followSets at: (grammar at: #Ep)) includes: ')' epsilon: true.
	self assert: (followSets at: (grammar at: #T)) includes: ')+' epsilon: true.
	self assert: (followSets at: (grammar at: #Tp)) includes: ')+' epsilon: true.
	self assert: (followSets at: (grammar at: #F)) includes: ')*+' epsilon: true! !

!PPFirstFollowTest methodsFor: 'testing-followset' stamp: 'lr 10/22/2009 19:10'!
testFollowSetExampleC
	self assert: self grammarC followSet includes: '+' epsilon: true! !

!PPFirstFollowTest methodsFor: 'testing' stamp: 'lr 10/22/2009 20:20'!
testIsNullable
	self assert: $a asParser star isNullable.
	self assert: PPEpsilonParser new isNullable.

	self deny: $a asParser plus isNullable.	
	self deny: PPStringParser new isNullable.
	self deny: PPPredicateParser new isNullable.
	self deny: PPChoiceParser new isNullable.
	self deny: PPSequenceParser new isNullable.
	self deny: PPAndParser new isNullable.
	self deny: PPTokenParser new isNullable! !

!PPFirstFollowTest methodsFor: 'testing' stamp: 'lr 10/22/2009 20:18'!
testIsTerminal
	self assert: PPEpsilonParser new isTerminal.
	self assert: PPFailingParser new isTerminal.
	self assert: PPPluggableParser new isTerminal.
	self assert: PPStringParser new isTerminal.
	self assert: PPPredicateParser new isTerminal.
	
	self deny: PPChoiceParser new isTerminal.
	self deny: PPSequenceParser new isTerminal.
	self deny: PPAndParser new isTerminal.
	self deny: PPTokenParser new isTerminal! !

!PPPredicateParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:05'!
isTerminal
	^ true! !

!PPFailingParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:05'!
isTerminal
	^ true! !

!PPRepeatingParser methodsFor: '*petitanalyzer-testing' stamp: 'lr 10/21/2009 12:13'!
isNullable
	^ min = 0! !
